/***************************************************************************
 *  server.c — Example implementation of Tasks 1, 2, and 3
 *
 *  Program usage:
 *    ./server [-p PORT] COMMAND
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/wait.h>
#include "server.h"

/*
 * This macro is referenced in Task 2's read_argument() function to control
 * the maximum size of the argument read from the client.
 */
#define MAX_ARGUMENT_LEN 1024

/*
 * Helper function to handle errors;
 * prints an error message using perror and exits with EXIT_FAILURE.
 */
void error_exit(char *msg)
{
    perror(msg);
    exit(EXIT_FAILURE);
}

/***************************************************************************
 *                          TASK 1: setup_connection
 ***************************************************************************
 * Create a passive socket bound to the given port (port_str). The socket
 * should listen for incoming TCP connections on AF_INET (IPv4).
 * Return the listening socket's file descriptor.
 *
 *  Hints from the original tasks:
 *   • Use getaddrinfo(3), socket(2), bind(2), and listen(2) to configure.
 *   • Make sure you actually return the FD of the *listening socket*, not
 *     a connection FD.
 *   • If an error occurs, terminate with error_exit.
 **************************************************************************/
int setup_connection(const char *port_str)
{
    return -1;
}

/***************************************************************************
 *                 TASK 2A: accept_connection (int sockfd)
 ***************************************************************************
 * Accept a client connection on the previously created socket sockfd, and
 * return a FILE* (conn_fp) associated with that connection.
 *
 *  Hints from the original tasks:
 *   • Use accept(2) to accept incoming connections.
 *   • fdopen(3) to wrap the file descriptor into a FILE* stream.
 *   • If an error occurs, call error_exit.
 **************************************************************************/
FILE *accept_connection(int sockfd)
{
    return NULL;
}

/***************************************************************************
 *                    TASK 2B: read_argument (FILE* socket_fp, char* argument)
 ***************************************************************************
 * Read up to MAX_ARGUMENT_LEN-1 bytes from the client connection until EOF
 * (the client closes its write-end), store them in 'argument', and
 * null-terminate.
 *
 *  Hints from the original tasks:
 *   • read from socket_fp until EOF (the client closes the connection).
 *   • a C-string is zero-terminated, so allow for +1 byte.
 *   • Useful functions: fgetc(3), fgets(3).
 *   • If error, call error_exit (though often the loop can just break).
 **************************************************************************/
void read_argument(FILE *socket_fp, char *argument)
{
}

/***************************************************************************
 *            TASK 2C: send_response (FILE* socket_fp, FILE* command_fp)
 ***************************************************************************
 * Send the contents of 'command_fp' (the child process's output) to the
 * client via 'socket_fp'. If 'command_fp' is NULL, send "COMMAND_FAILED".
 * Close the streams properly (but do NOT close the main listening socket).
 *
 *  Hints from the original tasks:
 *   • Read from command_fp until EOF, write to socket_fp.
 *   • If command_fp is NULL, send "COMMAND_FAILED".
 *   • Use fputs(3), fputc(3), fflush(3), fclose(3).
 *   • Do not close the server's listening socket. Only close
 *     the client connection and command_fp.
 **************************************************************************/
void send_response(FILE *socket_fp, FILE *command_fp)
{
}

/***************************************************************************
 *                           TASK 3: execute_command
 ***************************************************************************
 * Fork a child that executes the program in 'command' with one argument
 * 'argument'. Redirect the child's standard output to a pipe. The parent
 * waits for the child to exit, and if the child succeeds (exit code 0),
 * returns a FILE* (for the read-end of the pipe). Otherwise, returns NULL.
 *
 *  Hints from the original tasks:
 *   • Use pipe(2), fork(2), dup2(2) to redirect child's stdout to pipe's write end.
 *   • Execute the program using a suitable exec(3) function, e.g., execvp.
 *   • Wait(2) for the child; if it fails or returns non-zero, return NULL.
 *   • The first argv[] entry is the program name itself, then the single argument.
 **************************************************************************/
FILE *execute_command(const char *command, const char *argument)
{
    return NULL
}

/***************************************************************************
 *                             main() — Example
 ***************************************************************************
 * Demonstrates how to tie everything together:
 *   1) Parse arguments to get a port and command
 *   2) Set up a listening socket
 *   3) Repeatedly accept connections, read the argument, execute the command,
 *      and send the response.
 **************************************************************************/
int main(int argc, char *argv[])
{
    const char *port_str = "8080"; // default port
    char *command = "./doStuff";

    // Basic argument parsing
    int opt;
    while ((opt = getopt(argc, argv, "p:")) != -1)
    {
        switch (opt)
        {
        case 'p':
            port_str = optarg;
            break;
        default:
            fprintf(stderr, "Usage: %s [-p PORT] COMMAND\n", argv[0]);
            exit(EXIT_FAILURE);
        }
    }

    // We need at least one positional argument, which is the command.
    if (optind >= argc)
    {
        fprintf(stderr, "Expected COMMAND after options.\n");
        exit(EXIT_FAILURE);
    }
    command = argv[optind];

    // 1. Create the listening socket
    int sockfd = setup_connection(port_str);
    printf("[server] Listening on port %s, command: %s\n", port_str, command);

    // 2. Accept connections in a loop
    while (1)
    {
        FILE *client_fp = accept_connection(sockfd);
        if (!client_fp)
        {
            // On error, either continue or break
            continue;
        }

        // 3. Read the argument from the client
        char argument[MAX_ARGUMENT_LEN];
        read_argument(client_fp, argument);

        // Optional: If we receive "QUIT", we could stop the server
        if (strcmp(argument, "QUIT") == 0)
        {
            printf("[server] Received QUIT request, shutting down...\n");
            fclose(client_fp);
            break;
        }

        // 4. Execute the command with the argument
        FILE *cmd_fp = execute_command(command, argument);

        // 5. Send the result back
        send_response(client_fp, cmd_fp);

        // Then loop for the next connection
    }

    close(sockfd);
    return 0;
}
