/***************************************************************************
 *  server.c — Example implementation of Tasks 1, 2, and 3
 *
 *  Program usage:
 *    ./server [-p PORT] COMMAND
 *
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/wait.h>
#include "server.h"

/*
 * This macro is referenced in Task 2's read_argument() function to control
 * the maximum size of the argument read from the client.
 */
#define MAX_ARGUMENT_LEN 1024

/*
 * Helper function to handle errors;
 * prints an error message using perror and exits with EXIT_FAILURE.
 */
void error_exit(char *msg)
{
    perror(msg);
    exit(EXIT_FAILURE);
}

/***************************************************************************
 *                          TASK 1: setup_connection
 ***************************************************************************
 * Create a passive socket bound to the given port (port_str). The socket
 * should listen for incoming TCP connections on AF_INET (IPv4).
 * Return the listening socket's file descriptor.
 *
 *  Hints from the original tasks:
 *   • Use getaddrinfo(3), socket(2), bind(2), and listen(2) to configure.
 *   • Make sure you actually return the FD of the *listening socket*, not
 *     a connection FD.
 *   • If an error occurs, terminate with error_exit.
 **************************************************************************/
int setup_connection(const char *port_str)
{
    struct addrinfo hints, *ai;
    int sockfd;

    memset(&hints, 0, sizeof(hints));
    hints.ai_family = AF_INET;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_PASSIVE;

    if (getaddrinfo(NULL, port_str, &hints, &ai) != 0)
    {
        error_exit("getaddrinfo");
    }

    sockfd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
    if (sockfd == -1)
    {
        freeaddrinfo(ai);
        error_exit("socket");
    }

    if (bind(sockfd, ai->ai_addr, ai->ai_addrlen) == -1)
    {
        close(sockfd);
        freeaddrinfo(ai);
        error_exit("bind");
    }

    freeaddrinfo(ai);

    if (listen(sockfd, 1) == -1)
    {
        close(sockfd);
        error_exit("listen");
    }

    return sockfd;
}

/***************************************************************************
 *                 TASK 2A: accept_connection (int sockfd)
 ***************************************************************************
 * Accept a client connection on the previously created socket sockfd, and
 * return a FILE* (conn_fp) associated with that connection.
 *
 *  Hints from the original tasks:
 *   • Use accept(2) to accept incoming connections.
 *   • fdopen(3) to wrap the file descriptor into a FILE* stream.
 *   • If an error occurs, call error_exit.
 **************************************************************************/
FILE *accept_connection(int sockfd)
{
    int client_fd;

    client_fd = accept(sockfd, NULL, NULL);
    if (client_fd == -1)
    {
        error_exit("accept");
    }

    // Oder einfach return fdopen(client_fd, "r+"); geht hier anscheinend auch

    FILE *client_fp = fdopen(client_fd, "r+");
    if (client_fp == NULL)
    {
        close(client_fd);
        error_exit("fdopen");
    }

    return client_fp;
}

/***************************************************************************
 *                    TASK 2B: read_argument (FILE* socket_fp, char* argument)
 ***************************************************************************
 * Read up to MAX_ARGUMENT_LEN-1 bytes from the client connection until EOF
 * (the client closes its write-end), store them in 'argument', and
 * null-terminate.
 *
 *  Hints from the original tasks:
 *   • read from socket_fp until EOF (the client closes the connection).
 *   • a C-string is zero-terminated, so allow for +1 byte.
 *   • Useful functions: fgetc(3), fgets(3).
 *   • If error, call error_exit (though often the loop can just break).
 **************************************************************************/
void read_argument(FILE *socket_fp, char *argument)
{
    int c;
    size_t i = 0;
    while ((c = fgetc(socket_fp)) != EOF && i < MAX_ARGUMENT_LEN - 1)
    {
        argument[i++] = c;
    }
    argument[i] = '\0';
}

/***************************************************************************
 *            TASK 2C: send_response (FILE* socket_fp, FILE* command_fp)
 ***************************************************************************
 * Send the contents of 'command_fp' (the child process's output) to the
 * client via 'socket_fp'. If 'command_fp' is NULL, send "COMMAND_FAILED".
 * Close the streams properly (but do NOT close the main listening socket).
 *
 *  Hints from the original tasks:
 *   • Read from command_fp until EOF, write to socket_fp.
 *   • If command_fp is NULL, send "COMMAND_FAILED".
 *   • Use fputs(3), fputc(3), fflush(3), fclose(3).
 *   • Do not close the server's listening socket. Only close
 *     the client connection and command_fp.
 **************************************************************************/
void send_response(FILE *socket_fp, FILE *command_fp)
{
    char buffer[1024];
    // Oder char buffer[MAX_ARGUMENT_LEN+1]

    if (command_fp == NULL)
    {
        fputs("COMMAND_FAILED", socket_fp);
    }
    else
    {
        while (fgets(buffer, sizeof(buffer), command_fp) != NULL)
        {
            fputs(buffer, socket_fp);
        }
        fclose(command_fp);
    }
    fflush(socket_fp);
    fclose(socket_fp);
}

/***************************************************************************
 *                           TASK 3: execute_command
 ***************************************************************************
 * Fork a child that executes the program in 'command' with one argument
 * 'argument'. Redirect the child's standard output to a pipe. The parent
 * waits for the child to exit, and if the child succeeds (exit code 0),
 * returns a FILE* (for the read-end of the pipe). Otherwise, returns NULL.
 *
 *  Hints from the original tasks:
 *   • Use pipe(2), fork(2), dup2(2) to redirect child's stdout to pipe's write end.
 *   • Execute the program using a suitable exec(3) function, e.g., execvp.
 *   • Wait(2) for the child; if it fails or returns non-zero, return NULL.
 *   • The first argv[] entry is the program name itself, then the single argument.
 **************************************************************************/
FILE *execute_command(const char *command, const char *argument)
{
    int pipefd[2];
    if (pipe(pipefd) == -1)
    {
        error_exit("pipe");
    }

    pid_t pid = fork();
    if (pid == -1)
    {
        close(pipefd[0]);
        close(pipefd[1]);
        error_exit("fork");
    }

    if (pid == 0)
    { // Child process
        close(pipefd[0]);
        dup2(pipefd[1], STDOUT_FILENO);
        close(pipefd[1]);
        execlp(command, command, argument, NULL);
        error_exit("execlp"); // Bzw. exit(EXIT_FAILURE)
    }

    // Parent process
    close(pipefd[1]);

    int status;
    waitpid(pid, &status, 0);
    if (WEXITSTATUS(status) != EXIT_SUCCESS)
    {
        close(pipefd[0]);
        return NULL;
    }

    /* Oder einfach (anscheind auch möglich)
        if (wait(NULL) == -1) {
            close(pipefd[0]);
            error_exit("wait failed");
        }
     */

    FILE *command_fp = fdopen(pipefd[0], "r");
    if (command_fp == NULL)
    {
        close(pipefd[0]);
        error_exit("fdopen failed");
    }

    return command_fp;

    // Oder einfach return fdopen(pipefd[0], "r");
}

/***************************************************************************
 *                             main() — Example
 ***************************************************************************
 * Demonstrates how to tie everything together:
 *   1) Parse arguments to get a port and command
 *   2) Set up a listening socket
 *   3) Repeatedly accept connections, read the argument, execute the command,
 *      and send the response.
 **************************************************************************/
int main(int argc, char *argv[])
{
    const char *port_str = "8080"; // default port
    char *command = "./doStuff";

    // Basic argument parsing
    int opt;
    while ((opt = getopt(argc, argv, "p:")) != -1)
    {
        switch (opt)
        {
        case 'p':
            port_str = optarg;
            break;
        default:
            fprintf(stderr, "Usage: %s [-p PORT] COMMAND\n", argv[0]);
            exit(EXIT_FAILURE);
        }
    }

    // We need at least one positional argument, which is the command.
    if (optind >= argc)
    {
        fprintf(stderr, "Expected COMMAND after options.\n");
        exit(EXIT_FAILURE);
    }
    command = argv[optind];

    // 1. Create the listening socket
    int sockfd = setup_connection(port_str);
    printf("[server] Listening on port %s, command: %s\n", port_str, command);

    // 2. Accept connections in a loop
    while (1)
    {
        FILE *client_fp = accept_connection(sockfd);
        if (!client_fp)
        {
            // On error, either continue or break
            continue;
        }

        // 3. Read the argument from the client
        char argument[MAX_ARGUMENT_LEN];
        read_argument(client_fp, argument);

        // Optional: If we receive "QUIT", we could stop the server
        if (strcmp(argument, "QUIT") == 0)
        {
            printf("[server] Received QUIT request, shutting down...\n");
            fclose(client_fp);
            break;
        }

        // 4. Execute the command with the argument
        FILE *cmd_fp = execute_command(command, argument);

        // 5. Send the result back
        send_response(client_fp, cmd_fp);

        // Then loop for the next connection
    }

    close(sockfd);
    return 0;
}
